@@ -1,3 +1,8 @@
+"""
+A drop-in replacement for `argparse` that allows options to also be set via config files and/or environment variables.
+
+:see: `configargparse.ArgumentParser`, `configargparse.add_argument`
+"""
 import argparse
 import json
 import glob
@@ -14,9 +19,14 @@ else:
     from StringIO import StringIO
 
 
-ACTION_TYPES_THAT_DONT_NEED_A_VALUE = (argparse._StoreTrueAction,
+ACTION_TYPES_THAT_DONT_NEED_A_VALUE = [argparse._StoreTrueAction,
     argparse._StoreFalseAction, argparse._CountAction,
-    argparse._StoreConstAction, argparse._AppendConstAction)
+    argparse._StoreConstAction, argparse._AppendConstAction]
+
+if sys.version_info >= (3, 9):
+    ACTION_TYPES_THAT_DONT_NEED_A_VALUE.append(argparse.BooleanOptionalAction)
+
+ACTION_TYPES_THAT_DONT_NEED_A_VALUE = tuple(ACTION_TYPES_THAT_DONT_NEED_A_VALUE)
 
 
 # global ArgumentParser instances
@@ -80,7 +90,7 @@ class ConfigFileParser(object):
         one of: "yes", "no", "true", "false". Otherwise an error will be raised.
 
         Args:
-            stream: A config file input stream (such as an open file object).
+            stream (IO): A config file input stream (such as an open file object).
 
         Returns:
             OrderedDict: Items where the keys are strings and the
@@ -109,9 +119,11 @@ class ConfigFileParserException(Exception):
 
 
 class DefaultConfigFileParser(ConfigFileParser):
-    """Based on a simplified subset of INI and YAML formats. Here is the
-    supported syntax::
+    """
+    Based on a simplified subset of INI and YAML formats. Here is the
+    supported syntax
 
+    .. code::
 
         # this is a comment
         ; this is also a comment (.ini style)
@@ -142,7 +154,7 @@ class DefaultConfigFileParser(ConfigFileParser):
         return msg
 
     def parse(self, stream):
-        """Parses the keys + values from a config file."""
+       # see ConfigFileParser.parse docstring
 
         items = OrderedDict()
         for i, line in enumerate(stream):
@@ -178,9 +190,7 @@ class DefaultConfigFileParser(ConfigFileParser):
         return items
 
     def serialize(self, items):
-        """Does the inverse of config parsing by taking parsed values and
-        converting them back to a string representing config file contents.
-        """
+        # see ConfigFileParser.serialize docstring
         r = StringIO()
         for key, value in items.items():
             if isinstance(value, list):
@@ -211,7 +221,7 @@ class ConfigparserConfigFileParser(ConfigFileParser):
         return msg
 
     def parse(self, stream):
-        """Parses the keys and values from an INI config file."""
+        # see ConfigFileParser.parse docstring
         import configparser
         from ast import literal_eval
         # parse with configparser to allow multi-line values
@@ -246,9 +256,7 @@ class ConfigparserConfigFileParser(ConfigFileParser):
         return result
 
     def serialize(self, items):
-        """Does the inverse of config parsing by taking parsed values and
-        converting them back to a string representing config file contents.
-        """
+        # see ConfigFileParser.serialize docstring
         import configparser
         import io
         config = configparser.ConfigParser(
@@ -287,7 +295,7 @@ class YAMLConfigFileParser(ConfigFileParser):
         return yaml
 
     def parse(self, stream):
-        """Parses the keys and values from a config file."""
+        # see ConfigFileParser.parse docstring
         yaml = self._load_yaml()
 
         try:
@@ -313,12 +321,7 @@ class YAMLConfigFileParser(ConfigFileParser):
         return result
 
     def serialize(self, items, default_flow_style=False):
-        """Does the inverse of config parsing by taking parsed values and
-        converting them back to a string representing config file contents.
-
-        Args:
-            default_flow_style: defines serialization format (see PyYAML docs)
-        """
+        # see ConfigFileParser.serialize docstring
 
         # lazy-import so there's no dependency on yaml unless this class is used
         yaml = self._load_yaml()
@@ -336,13 +339,13 @@ _DEFAULTS_SOURCE_KEY = "defaults"
 
 
 class ArgumentParser(argparse.ArgumentParser):
-    """Drop-in replacement for argparse.ArgumentParser that adds support for
-    environment variables and .ini or .yaml-style config files.
+    """Drop-in replacement for `argparse.ArgumentParser` that adds support for
+    environment variables and ``.ini`` or ``.yaml-style`` config files.
     """
 
     def __init__(self, *args, **kwargs):
 
-        r"""Supports args of the argparse.ArgumentParser constructor
+        r"""Supports args of the `argparse.ArgumentParser` constructor
         as \*\*kwargs, as well as the following additional args.
 
         Arguments:
@@ -353,8 +356,8 @@ class ArgumentParser(argparse.ArgumentParser):
             auto_env_var_prefix: If set to a string instead of None, all config-
                 file-settable options will become also settable via environment
                 variables whose names are this prefix followed by the config
-                file key, all in upper case. (eg. setting this to "foo_" will
-                allow an arg like "--my-arg" to also be set via the FOO_MY_ARG
+                file key, all in upper case. (eg. setting this to ``foo_`` will
+                allow an arg like ``--my-arg`` to also be set via the FOO_MY_ARG
                 environment variable)
             default_config_files: When specified, this list of config files will
                 be parsed in order, with the values from each config file
@@ -448,13 +451,16 @@ class ArgumentParser(argparse.ArgumentParser):
 
     def parse_args(self, args = None, namespace = None,
                    config_file_contents = None, env_vars = os.environ):
-        """Supports all the same args as the ArgumentParser.parse_args(..),
+        """Supports all the same args as the `argparse.ArgumentParser.parse_args()`,
         as well as the following additional args.
 
         Arguments:
             args: a list of args as in argparse, or a string (eg. "-x -y bla")
             config_file_contents: String. Used for testing.
             env_vars: Dictionary. Used for testing.
+        
+        Returns:
+            argparse.Namespace: namespace
         """
         args, argv = self.parse_known_args(
             args=args,
@@ -475,15 +481,18 @@ class ArgumentParser(argparse.ArgumentParser):
             env_vars=os.environ,
             ignore_help_args=False,
     ):
-        """Supports all the same args as the ArgumentParser.parse_args(..),
+        """Supports all the same args as the `argparse.ArgumentParser.parse_args()`,
         as well as the following additional args.
 
         Arguments:
             args: a list of args as in argparse, or a string (eg. "-x -y bla")
             config_file_contents (str). Used for testing.
             env_vars (dict). Used for testing.
-            ignore_help_args (bool): This flag determines behavior when user specifies --help or -h. If False,
+            ignore_help_args (bool): This flag determines behavior when user specifies ``--help`` or ``-h``. If False,
                 it will have the default behavior - printing help and exiting. If True, it won't do either.
+        
+        Returns:
+            tuple[argparse.Namespace, list[str]]: tuple namescpace, unknown_args
         """
         if args is None:
             args = sys.argv[1:]
@@ -532,9 +541,13 @@ class ArgumentParser(argparse.ArgumentParser):
             # Make list-string into list.
             if action.nargs or isinstance(action, argparse._AppendAction):
                 nargs = True
-                element_capture = re.match(r'\[(.*)\]', value)
-                if element_capture:
-                    value = [val.strip() for val in element_capture.group(1).split(',') if val.strip()]
+                if value.startswith("[") and value.endswith("]"):
+                    # handle special case of k=[1,2,3] or other json-like syntax
+                    try:
+                        value = json.loads(value)
+                    except Exception:
+                        # for backward compatibility with legacy format (eg. where config value is [a, b, c] instead of proper json ["a", "b", "c"]
+                        value = [elem.strip() for elem in value[1:-1].split(",")]
             env_var_args += self.convert_item_to_command_line_arg(
                 action, key, value)
 
@@ -642,14 +655,18 @@ class ArgumentParser(argparse.ArgumentParser):
         return namespace, unknown_args
 
     def get_source_to_settings_dict(self):
-        """If called after parse_args() or parse_known_args(), this dict. will contain up to 4 keys corresponding
+        """
+        If called after `parse_args()` or `parse_known_args()`, returns a dict that contains up to 4 keys corresponding
         to where a given option's value is coming from:
-            "command_line"
-            "environment_variables"
-            "config_file"
-            "defaults"
+        - "command_line"
+        - "environment_variables"
+        - "config_file"
+        - "defaults"
         Each such key, will be mapped to another dictionary containing the options set via that method. Here the key
-        will be the option name, and the value will be a 2-tuple of the form (argparse Action obj, string value).
+        will be the option name, and the value will be a 2-tuple of the form (`argparse.Action` obj, `str` value).
+
+        Returns:
+            dict[str, dict[str, tuple[argparse.Action, str]]]: source to settings dict
         """
 
         return self._source_to_settings
@@ -691,6 +708,9 @@ class ArgumentParser(argparse.ArgumentParser):
 
         Args:
             key: The config file key that was being set.
+        
+        Returns:
+            str: command line key
         """
         key_without_prefix_chars = key.strip(self.prefix_chars)
         command_line_key = self.prefix_chars[0]*2 + key_without_prefix_chars
@@ -754,6 +774,9 @@ class ArgumentParser(argparse.ArgumentParser):
                 configargparse arg.
             key: string (config file key or env var name)
             value: parsed value of type string or list
+        
+        Returns:
+            list[str]: args
         """
         args = []
 
@@ -812,6 +835,9 @@ class ArgumentParser(argparse.ArgumentParser):
         """This method decides which actions can be set in a config file and
         what their keys will be. It returns a list of 0 or more config keys that
         can be used to set the given action's value in a config file.
+
+        Returns:
+            list[str]: keys
         """
         keys = []
 
@@ -833,6 +859,9 @@ class ArgumentParser(argparse.ArgumentParser):
 
         Args:
             command_line_args: List of all args
+        
+        Returns:
+            list[IO]: open config files
         """
         # open any default config files
         config_files = []
@@ -900,6 +929,9 @@ class ArgumentParser(argparse.ArgumentParser):
     def format_values(self):
         """Returns a string with all args and settings and where they came from
         (eg. commandline, config file, environment variable or default)
+
+        Returns:
+            str: source to settings string
         """
         source_key_to_display_value_map = {
             _COMMAND_LINE_SOURCE_KEY: "Command Line Args: ",
@@ -1009,6 +1041,9 @@ def add_argument(self, *args, **kwargs):
             configargparse to write all current commandline args to this file
             as config options and then exit.
             Default: False
+    
+    Returns:
+        argparse.Action: the new argparse action
     """
 
     env_var = kwargs.pop("env_var", None)
@@ -1046,6 +1081,9 @@ def add_argument(self, *args, **kwargs):
 def already_on_command_line(existing_args_list, potential_command_line_args, prefix_chars):
     """Utility method for checking if any of the potential_command_line_args is
     already present in existing_args.
+
+    Returns:
+        bool: already on command line?
     """
     arg_names = []
     for arg_string in existing_args_list:
@@ -1058,7 +1096,8 @@ def already_on_command_line(existing_args_list, potential_command_line_args, pre
     return any(
         potential_arg in arg_names for potential_arg in potential_command_line_args
     )
-
+#TODO: Update to latest version of pydoctor when https://github.com/twisted/pydoctor/pull/414 has been merged 
+# such that the alises can be documented automatically.
 
 # wrap ArgumentParser's add_argument(..) method with the one above
 argparse._ActionsContainer.original_add_argument_method = argparse._ActionsContainer.add_argument
